# Demo: Robust PCA for background-foreground separation of videos
# Task:
#   We shall apply Alternating Direction Method of Multipliers (ADMM) to solve
# this model, where the augmented Lagrangian is given by:
#   L(A,B,M,Y) = |A|_nuc + lambda*|B|_l1 + delta{|M-Z|_fro <= eps} +
#   <Y, A+B-M> + 0.5*rho*|A+B-M|_fro^2
#
# #############################################################################


import os
import numpy as np
# import scipy.sparse.linalg as splinalg
import matplotlib.pyplot as plt
from PIL import Image


def im2arr(img):
    """pillow grayscale image to normalized numpy array"""
    return np.array(img, dtype=np.float64) / 255.0


def imnorm(imgmat):
    """normalize image from matrix representation"""
    imgmat = (imgmat - np.min(imgmat)) * 255. / \
        (np.max(imgmat) - np.min(imgmat) + 10e-20)
    return Image.fromarray(imgmat.astype(np.uint8)).convert('L')


def mat2vec(m):
    """flatten matrix m"""
    return m.T.flatten()


def vec2mat(v, shape):
    """unflatten vectorized matrix v"""
    h, w = shape
    return v.reshape(w, h).T


def get_synthetic_data(nFrame):
    # parameters
    h, w = 20, 25
    n1, n2 = h * w, nFrame       # matrix size
    r = min(n1, n2) // 10   # presumed matrix rank
    s = (n1 * n2) // 10     # presumed matrix sparsity
    noise = 0.01            # noise level
    # true low-rank matrix A_0
    A_0 = np.random.randn(n1, r) @ np.random.randn(r, n2)
    A_max = np.sqrt(max(n1, n2))
    # true sparse matrix B_0
    B_0 = ((np.random.rand(n1, n2) < 0.5) - 0.5) * 2 * A_max
    B_0[np.unravel_index(np.random.permutation(B_0.size)[:n1*n2-s],
                         B_0.shape)] = 0
    # build up the noise output
    Z = A_0 + B_0
    Z += noise * np.random.randn(*Z.shape)
    return Z, h, w


def get_airport_data(nFrame):
    # parameters
    data_dir = 'video_airport/'
    initFrame = 1000
    w, h = Image.open(os.path.join(data_dir, 'airport1000.bmp')).size
    # read data
    img_arrays = []
    for i in range(initFrame, initFrame + nFrame):
        img_name = os.path.join(data_dir, 'airport{}.bmp'.format(i))
        img = mat2vec(im2arr(Image.open(img_name).convert('L')))
        img_arrays.append(img)
    return np.stack(img_arrays, axis=1), h, w


def plot_augmented_lagrangians(energies):
    plt.figure('augmented lagrangian')
    plt.plot(range(len(energies)), energies)
    plt.show()


def display_input(Z, m, n):
    """display input video frame-by-frame"""
    plt.figure('Input video')
    n1, n2 = Z.shape
    for i in range(n2):
        plt.imshow(imnorm(vec2mat(Z[:, i], (m, n))))
        plt.title('original ({})'.format(i))
        plt.draw()
        plt.pause(1.0 / 12)  # show at 12 fps
        plt.clf()
    plt.close()


def display_output(A, B, Z, m, n):
    """display output frame-by-frame showing the decomposition"""
    plt.figure('Foreground/background separation')
    n1, n2 = Z.shape
    for i in range(n2):
        plt.imshow(imnorm(np.concatenate(
                [vec2mat(A[:, i], (m, n)), vec2mat(B[:, i], (m, n)),
                 vec2mat(Z[:, i], (m, n))], axis=1)))
        plt.title('background / foreground / original ({})'.format(i))
        plt.draw()
        plt.pause(1.0 / 12)  # show at 12 fps
        plt.clf()


if __name__ == '__main__':
    # get input data
    use_synthetic_data = True
    nFrame = 20  # should be <= 200 for the real airport data
    Z, m, n = (get_airport_data, get_synthetic_data
               )[use_synthetic_data](nFrame)
    n1, n2 = Z.shape

    # display input data
    display_input(Z, m, n)

    # algorithm parameters
    lamb = 0.5 / np.sqrt(n1)
    eps = 0.01
    rho = 1.0
    max_it = 100

    # TODO: initialization of A, B, M, Y
    A = NotImplemented
    B = NotImplemented
    energies = []

    raise NotImplementedError

    # TODO: ADMM solver
    for it in range(max_it):
        # update A:

        # update B:

        # update M:

        # update Y:

        # update augmented lagrangian energies

        # output status

        raise NotImplementedError

    # visualize results
    plot_augmented_lagrangians(energies)
    display_output(A, B, Z, m, n)
